from importlib import metadata

from langchain_arangodb.chains.graph_qa.arangodb import ArangoGraphQAChain
from langchain_arangodb.chat_message_histories.arangodb import ArangoChatMessageHistory
from langchain_arangodb.graphs.arangodb_graph import ArangoGraph
from langchain_arangodb.vectorstores.arangodb_vector import ArangoVector

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""
del metadata  # optional, avoids polluting the results of dir(__package__)

__all__ = [
    "ArangoGraphQAChain",
    "ArangoChatMessageHistory",
    "ArangoGraph",
    "ArangoVector",
    "__version__",
]



from __future__ import annotations

from enum import Enum
from typing import Any, Callable, Iterable, List, Optional, Sequence, Tuple, Type, Union

import farmhash
import numpy as np
from arango.aql import Cursor
from arango.database import StandardDatabase
from arango.exceptions import ArangoServerError, ViewGetError
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore
from langchain_core.vectorstores.utils import maximal_marginal_relevance
from packaging import version

from langchain_arangodb.vectorstores.utils import DistanceStrategy

DEFAULT_DISTANCE_STRATEGY = DistanceStrategy.COSINE
DISTANCE_MAPPING = {
    DistanceStrategy.EUCLIDEAN_DISTANCE: "l2",
    DistanceStrategy.COSINE: "cosine",
}


class SearchType(str, Enum):
    """Enumerator of the search types."""

    VECTOR = "vector"
    HYBRID = "hybrid"


DEFAULT_SEARCH_TYPE = SearchType.VECTOR

# Constants for RRF
DEFAULT_RRF_CONSTANT = 60  # Standard constant for RRF
DEFAULT_SEARCH_LIMIT = 100  # Default limit for initial search results

# Full-text search analyzer options
DEFAULT_ANALYZER = "text_en"  # Default analyzer for full-text search
SUPPORTED_ANALYZERS = [
    "text_en",
    "text_de",
    "text_es",
    "text_fi",
    "text_fr",
    "text_it",
    "text_nl",
    "text_no",
    "text_pt",
    "text_ru",
    "text_sv",
    "text_zh",
]


class ArangoVector(VectorStore):
    """ArangoDB vector store implementation for LangChain.

    This class provides a vector store implementation using ArangoDB as the backend.
    It supports both vector similarity search and hybrid search (vector + keyword)
    capabilities.

    :param embedding: The embedding function to use for converting text to vectors.
        Must implement the `langchain.embeddings.base.Embeddings` interface.
    :type embedding: langchain.embeddings.base.Embeddings
    :param embedding_dimension: The dimensionality of the embedding vectors.
        Must match the output dimension of the embedding function.
    :type embedding_dimension: int
    :param database: The ArangoDB database instance to use for storage and retrieval.
    :type database: arango.database.StandardDatabase
    :param collection_name: The name of the ArangoDB collection to store
        documents. Defaults to "documents".
    :type collection_name: str
    :param search_type: The type of search to perform. Can be either SearchType.VECTOR
        for pure vector similarity search or SearchType.HYBRID for combining vector and
        keyword search. Defaults to SearchType.VECTOR.
    :type search_type: SearchType
    :param embedding_field: The field name in the document to store the embedding vector
        Defaults to "embedding".
    :type embedding_field: str
    :param text_field: The field name in the document to store the text content.
        Defaults to "text".
    :type text_field: str
    :param vector_index_name: The name of the vector index to create in ArangoDB.
        This index enables efficient vector similarity search.
        Defaults to "vector_index".
    :type vector_index_name: str
    :param distance_strategy: The distance metric to use for vector similarity.
        Can be either DistanceStrategy.COSINE or DistanceStrategy.EUCLIDEAN_DISTANCE.
        Defaults to DistanceStrategy.COSINE.
    :type distance_strategy: DistanceStrategy
    :param num_centroids: The number of centroids to use for the vector index.
        Higher values can improve search accuracy but increase memory usage.
        Defaults to 1.
    :type num_centroids: int
    :param relevance_score_fn: Optional function to normalize the relevance score.
        If not provided, uses the default normalization for the distance strategy.
    :type relevance_score_fn: Optional[Callable[[float], float]]
    :param keyword_index_name: The name of the ArangoDB View created to enable
        Full-Text-Search capabilities. Only used if search_type is set
        to SearchType.HYBRID. Defaults to "keyword_index".
    :type keyword_index_name: str
    :param keyword_analyzer: The text analyzer to use for keyword search.
        Must be one of the supported analyzers in ArangoDB.
        Defaults to "text_en".
    :type keyword_analyzer: str
    :param rrf_constant: The constant used in Reciprocal Rank Fusion (RRF) for hybrid
        search. Higher values give more weight to lower-ranked results.
        Defaults to 60.
    :type rrf_constant: int
    :param rrf_search_limit: The maximum number of results to consider in RRF scoring.
        Defaults to 100.
    :type rrf_search_limit: int
    """

    def __init__(
        self,
        embedding: Embeddings,
        embedding_dimension: int,
        database: StandardDatabase,
        collection_name: str = "documents",
        search_type: SearchType = DEFAULT_SEARCH_TYPE,
        embedding_field: str = "embedding",
        text_field: str = "text",
        vector_index_name: str = "vector_index",
        distance_strategy: DistanceStrategy = DEFAULT_DISTANCE_STRATEGY,
        num_centroids: int = 1,
        relevance_score_fn: Optional[Callable[[float], float]] = None,
        keyword_index_name: str = "keyword_index",
        keyword_analyzer: str = DEFAULT_ANALYZER,
        rrf_constant: int = DEFAULT_RRF_CONSTANT,
        rrf_search_limit: int = DEFAULT_SEARCH_LIMIT,
    ):
        if search_type not in [SearchType.VECTOR, SearchType.HYBRID]:
            raise ValueError("search_type must be 'vector' or 'hybrid'")

        if distance_strategy not in [
            DistanceStrategy.COSINE,
            DistanceStrategy.EUCLIDEAN_DISTANCE,
        ]:
            m = "distance_strategy must be 'COSINE' or 'EUCLIDEAN_DISTANCE'"
            raise ValueError(m)

        self.embedding = embedding
        self.embedding_dimension = int(embedding_dimension)
        self.db = database
        self.async_db = self.db.begin_async_execution(return_result=False)
        self.search_type = search_type
        self.collection_name = collection_name
        self.embedding_field = embedding_field
        self.text_field = text_field
        self.vector_index_name = vector_index_name
        self._distance_strategy = distance_strategy
        self.num_centroids = num_centroids
        self.override_relevance_score_fn = relevance_score_fn

        # Hybrid search parameters
        self.keyword_index_name = keyword_index_name
        self.keyword_analyzer = keyword_analyzer
        self.rrf_constant = rrf_constant
        self.rrf_search_limit = rrf_search_limit

        if not self.db.has_collection(collection_name):
            self.db.create_collection(collection_name)

        self.collection = self.db.collection(self.collection_name)

    @property
    def embeddings(self) -> Embeddings:
        return self.embedding

    def retrieve_vector_index(self) -> Union[dict[str, Any], None]:
        """Retrieve the vector index from the collection."""
        indexes = self.collection.indexes()  # type: ignore
        for index in indexes:  # type: ignore
            if index["name"] == self.vector_index_name:
                return index

        return None

    def create_vector_index(self) -> None:
        """Create the vector index on the collection."""
        self.collection.add_index(  # type: ignore
            {
                "name": self.vector_index_name,
                "type": "vector",
                "fields": [self.embedding_field],
                "params": {
                    "metric": DISTANCE_MAPPING[self._distance_strategy],
                    "dimension": self.embedding_dimension,
                    "nLists": self.num_centroids,
                },
            }
        )

    def delete_vector_index(self) -> None:
        """Delete the vector index from the collection."""
        index = self.retrieve_vector_index()

        if index is not None:
            self.collection.delete_index(index["id"])

    def retrieve_keyword_index(self) -> Optional[dict[str, Any]]:
        """Retrieve the keyword index from the collection."""
        try:
            return self.db.view(self.keyword_index_name)  # type: ignore
        except ViewGetError:
            return None

    def create_keyword_index(self) -> None:
        """Create the keyword index on the collection."""
        if self.retrieve_keyword_index():
            return

        collection = self.db.collection(self.collection_name)
        collection.add_index(
            {
                "type": "inverted",
                "name": self.keyword_index_name,
                "fields": [
                    {"name": self.text_field, "analyzer": self.keyword_analyzer}
                ],
            }
        )
        view_properties = {
            "indexes": [
                {"collection": self.collection_name, "index": self.keyword_index_name}
            ]
        }
        self.db.create_view(self.keyword_index_name, "search-alias", view_properties)

    def delete_keyword_index(self) -> None:
        """Delete the keyword index from the collection."""
        view = self.retrieve_keyword_index()
        if view:
            self.db.delete_view(self.keyword_index_name)
            self.db.collection(self.collection_name).delete_index(
                self.keyword_index_name, ignore_missing=True
            )

    def add_embeddings(
        self,
        texts: Iterable[str],
        embeddings: List[List[float]],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        batch_size: int = 500,
        use_async_db: bool = False,
        insert_text: bool = True,
        **kwargs: Any,
    ) -> List[str]:
        """Add embeddings to the vectorstore."""
        texts = list(texts)

        if ids is None:
            ids = [str(farmhash.Fingerprint64(text.encode("utf-8"))) for text in texts]  # type: ignore

        if not metadatas:
            metadatas = [{} for _ in texts]

        if len(ids) != len(texts) != len(embeddings) != len(metadatas):
            m = "Length of ids, texts, embeddings and metadatas must be the same."
            raise ValueError(m)

        db = self.async_db if use_async_db else self.db
        collection = db.collection(self.collection_name)

        data = []
        for _key, text, embedding, metadata in zip(ids, texts, embeddings, metadatas):
            doc: dict[str, Any] = {self.text_field: text} if insert_text else {}

            doc.update(
                {
                    **metadata,
                    "_key": _key,
                    self.embedding_field: embedding,
                }
            )

            data.append(doc)

            if len(data) == batch_size:
                collection.import_bulk(data, on_duplicate="update", **kwargs)
                data = []

        collection.import_bulk(data, on_duplicate="update", **kwargs)

        return ids

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Add texts to the vector store.

        This method embeds the provided texts using the embedding function and stores
        them in ArangoDB along with their embeddings and metadata.

        :param texts: An iterable of text strings to add to the vector store.
        :type texts: Iterable[str]
        :param metadatas: Optional list of metadata dictionaries to associate with each
            text. Each dictionary can contain arbitrary key-value pairs that
            will be stored alongside the text and embedding.
        :type metadatas: Optional[List[dict]]
        :param ids: Optional list of unique identifiers for each text. If not provided,
            IDs will be generated using a hash of the text content.
        :type ids: Optional[List[str]]
        :param kwargs: Additional keyword arguments passed to add_embeddings.
        :type kwargs: Any
        :return: List of document IDs that were added to the vector store.
        :rtype: List[str]

        .. code-block:: python

            # Add simple texts
            texts = ["hello world", "hello arango", "test document"]
            ids = vector_store.add_texts(texts)
            print(f"Added {len(ids)} documents")

            # Add texts with metadata
            texts = ["Machine learning tutorial", "Python programming guide"]
            metadatas = [
                {"category": "AI", "difficulty": "beginner"},
                {"category": "Programming", "difficulty": "intermediate"}
            ]
            ids = vector_store.add_texts(texts, metadatas=metadatas)

            # Add texts with custom IDs
            texts = ["Document 1", "Document 2"]
            custom_ids = ["doc_001", "doc_002"]
            ids = vector_store.add_texts(texts, ids=custom_ids)
        """
        embeddings = self.embedding.embed_documents(list(texts))

        return self.add_embeddings(
            texts=texts, embeddings=embeddings, metadatas=metadatas, ids=ids, **kwargs
        )

    def similarity_search(
        self,
        query: str,
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        embedding: Optional[List[float]] = None,
        filter_clause: str = "",
        search_type: Optional[SearchType] = None,
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
        metadata_clause: str = "",
        **kwargs: Any,
    ) -> List[Document]:
        """Search for similar documents using vector similarity or hybrid search.

        This method performs a similarity search using either pure vector similarity
        or a hybrid approach combining vector and keyword search. The search type
        can be overridden for individual queries.

        :param query: The text query to search for.
        :type query: str
        :param k: The number of most similar documents to return. Defaults to 4.
        :type k: int
        :param return_fields: Set of additional document fields to return in results.
            The _key and text fields are always returned.
        :type return_fields: set[str]
        :param use_approx: Whether to use approximate nearest neighbor search.
            Enables faster but potentially less accurate results.
            Defaults to True.
        :type use_approx: bool
        :param embedding: Optional pre-computed embedding for the query.
            If not provided, the query will be embedded using the embedding
            function.
        :type embedding: Optional[List[float]]
        :param filter_clause: Optional AQL filter clause to apply to the search.
            Can be used to filter results based on document properties.
        :type filter_clause: str
        :param search_type: Override the default search type for this query.
            Can be either SearchType.VECTOR or SearchType.HYBRID.
        :type search_type: Optional[SearchType]
        :param vector_weight: Weight to apply to vector similarity
            scores in hybrid search. Only used when search_type is SearchType.HYBRID.
            Defaults to 1.0.
        :type vector_weight: float
        :param keyword_weight: Weight to apply to keyword search scores in hybrid
            search. Only used when search_type is SearchType.HYBRID. Defaults to 1.0.
        :type keyword_weight: float
        :param keyword_search_clause: Optional AQL filter clause to apply
          Full Text Search. If empty, a default search clause will be used.
        :type keyword_search_clause: str
        :param metadata_clause: Optional AQL clause to return additional metadata once
            the top k results are retrieved. If specified, the metadata will be
            added to the Document.metadata field.
        :type metadata_clause: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any
        :return: List of Document objects most similar to the query.
        :rtype: List[Document]

        .. code-block:: python

            # Simple vector search
            results = vector_store.similarity_search("hello", k=1)
            print(results[0].page_content)

            # Search with metadata filtering
            results = vector_store.similarity_search(
                "machine learning",
                k=2,
                filter_clause="doc.category == 'AI'",
                return_fields={"category", "difficulty"}
            )

            # Hybrid search with custom weights
            results = vector_store.similarity_search(
                "neural networks",
                k=3,
                search_type=SearchType.HYBRID,
                vector_weight=0.8,
                keyword_weight=0.2
            )
        """
        search_type = search_type or self.search_type
        embedding = embedding or self.embedding.embed_query(query)

        if search_type == SearchType.VECTOR:
            return self.similarity_search_by_vector(
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
                metadata_clause=metadata_clause,
            )

        else:
            return self.similarity_search_by_vector_and_keyword(
                query=query,
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
                vector_weight=vector_weight,
                keyword_weight=keyword_weight,
                keyword_search_clause=keyword_search_clause,
                metadata_clause=metadata_clause,
            )

    def similarity_search_with_score(
        self,
        query: str,
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        embedding: Optional[List[float]] = None,
        filter_clause: str = "",
        search_type: Optional[SearchType] = None,
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
        metadata_clause: str = "",
    ) -> List[tuple[Document, float]]:
        """Search for similar documents and return their similarity scores.

        Similar to similarity_search but returns a tuple of (Document, score) for each
        result. The score represents the similarity between the query and the document.

        :param query: The text query to search for.
        :type query: str
        :param k: The number of most similar documents to return. Defaults to 4.
        :type k: int
        :param return_fields: Set of additional document fields to return in results.
            The _key and text fields are always returned.
        :type return_fields: set[str]
        :param use_approx: Whether to use approximate nearest neighbor search.
            Enables faster but potentially less accurate results. Defaults to True.
        :type use_approx: bool
        :param embedding: Optional pre-computed embedding for the query.
            If not provided, the query will be embedded using the embedding function.
        :type embedding: Optional[List[float]]
        :param filter_clause: Optional AQL filter clause to apply to the search.
            Can be used to filter results based on document properties.
        :type filter_clause: str
        :param search_type: Override the default search type for this query.
            Can be either SearchType.VECTOR or SearchType.HYBRID.
        :type search_type: Optional[SearchType]
        :param vector_weight: Weight to apply to vector similarity scores in
            hybrid search. Only used when search_type is SearchType.HYBRID.
            Defaults to 1.0.
        :type vector_weight: float
        :param keyword_weight: Weight to apply to keyword search scores
            in hybrid search. Only used when search_type is SearchType.HYBRID.
            Defaults to 1.0.
        :type keyword_weight: float
        :param keyword_search_clause: Optional AQL filter clause to apply
            Full Text Search. If empty, a default search clause will be used.
        :type keyword_search_clause: str
        :param metadata_clause: Optional AQL clause to return additional metadata once
            the top k results are retrieved.
        :type metadata_clause: str
        :return: List of tuples containing (Document, score) pairs, sorted by score.
        :rtype: List[tuple[Document, float]]
        """
        search_type = search_type or self.search_type
        embedding = embedding or self.embedding.embed_query(query)

        if search_type == SearchType.VECTOR:
            return self.similarity_search_by_vector_with_score(
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
                metadata_clause=metadata_clause,
            )

        else:
            return self.similarity_search_by_vector_and_keyword_with_score(
                query=query,
                embedding=embedding,
                k=k,
                return_fields=return_fields,
                use_approx=use_approx,
                filter_clause=filter_clause,
                vector_weight=vector_weight,
                keyword_weight=keyword_weight,
                keyword_search_clause=keyword_search_clause,
                metadata_clause=metadata_clause,
            )

    def similarity_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        metadata_clause: str = "",
        **kwargs: Any,
    ) -> List[Document]:
        """Return docs most similar to embedding vector.

        :param embedding: Embedding to look up documents similar to.
        :type embedding: List[float]
        :param k: Number of Documents to return. Defaults to 4.
        :type k: int
        :param return_fields: Fields to return in the result. For example,
            {"foo", "bar"} will return the "foo" and "bar" fields of the document,
            in addition to the _key & text field. Defaults to an empty set.
        :type return_fields: set[str]
        :param use_approx: Whether to use approximate vector search via ANN.
            Defaults to True. If False, exact vector search will be used.
        :type use_approx: bool
        :param filter_clause: Filter clause to apply to the query.
        :type filter_clause: str
        :param metadata_clause: Optional AQL clause to return additional metadata once
            the top k results are retrieved. If specified, the metadata will be
            added to the Document.metadata field.
        :type metadata_clause: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any
        :return: List of Documents most similar to the query vector.
        :rtype: List[Document]
        """
        results = self.similarity_search_by_vector_with_score(
            embedding=embedding,
            k=k,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
            metadata_clause=metadata_clause,
        )

        return [doc for doc, _ in results]

    def similarity_search_by_vector_and_keyword(
        self,
        query: str,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
        metadata_clause: str = "",
    ) -> List[Document]:
        results = self.similarity_search_by_vector_and_keyword_with_score(
            query=query,
            embedding=embedding,
            k=k,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
            vector_weight=vector_weight,
            keyword_weight=keyword_weight,
            keyword_search_clause=keyword_search_clause,
            metadata_clause=metadata_clause,
        )

        return [doc for doc, _ in results]

    def similarity_search_by_vector_with_score(
        self,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        metadata_clause: str = "",
    ) -> List[tuple[Document, float]]:
        """Return docs most similar to embedding vector with scores.

        :param embedding: Embedding to look up documents similar to.
        :type embedding: List[float]
        :param k: Number of Documents to return. Defaults to 4.
        :type k: int
        :param return_fields: Fields to return in the result. For example,
            {"foo", "bar"} will return the "foo" and "bar" fields of the document,
            in addition to the _key & text field. Defaults to an empty set.
        :type return_fields: set[str]
        :param use_approx: Whether to use approximate vector search via ANN.
            Defaults to True. If False, exact vector search will be used.
        :type use_approx: bool
        :param filter_clause: Filter clause to apply to the query.
        :type filter_clause: str
        :param metadata_clause: Optional AQL clause to return additional metadata once
            the top k results are retrieved. If specified, the metadata will be
            added to the Document.metadata field.
        :type metadata_clause: str
        :return: List of tuples containing (Document, score)
            pairs most similar to the query vector.
        :rtype: List[tuple[Document, float]]
        """
        aql_query, bind_vars = self._build_vector_search_query(
            embedding=embedding,
            k=k,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
            metadata_clause=metadata_clause,
        )

        cursor = self.db.aql.execute(aql_query, bind_vars=bind_vars, stream=True)

        results = self._process_search_query(cursor)  # type: ignore

        return results

    def similarity_search_by_vector_and_keyword_with_score(
        self,
        query: str,
        embedding: List[float],
        k: int = 4,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        filter_clause: str = "",
        vector_weight: float = 1.0,
        keyword_weight: float = 1.0,
        keyword_search_clause: str = "",
        metadata_clause: str = "",
    ) -> List[tuple[Document, float]]:
        """Run hybrid similarity search combining vector and keyword search with scores.

        :param query: Query text to search for.
        :type query: str
        :param embedding: Embedding vector for the query.
        :type embedding: List[float]
        :param k: Number of results to return. Defaults to 4.
        :type k: int
        :param return_fields: Fields to return in the result. For example,
            {"foo", "bar"} will return the "foo" and "bar" fields of the document,
            in addition to the _key & text field. Defaults to an empty set.
        :type return_fields: set[str]
        :param use_approx: Whether to use approximate vector search via ANN.
            Defaults to True. If False, exact vector search will be used.
        :type use_approx: bool
        :param filter_clause: Filter clause to apply to the query.
        :type filter_clause: str
        :param vector_weight: Weight to apply to vector similarity scores
            in hybrid search. Only used when search_type is SearchType.HYBRID.
            Defaults to 1.0.
        :type vector_weight: float
        :param keyword_weight: Weight to apply to keyword search scores in
            hybrid search. Only used when search_type is SearchType.HYBRID.
            Defaults to 1.0.
        :type keyword_weight: float
        :param keyword_search_clause: Optional AQL filter clause to apply
            Full Text Search. If empty, a default search clause will be used.
        :type keyword_search_clause: str
        :param metadata_clause: Optional AQL clause to return additional metadata once
            the top k results are retrieved. If specified, the metadata will be
            added to the Document.metadata field.
        :type metadata_clause: str
        :return: List of tuples containing (Document, score)
            pairs most similar to thequery.
        :rtype: List[tuple[Document, float]]
        """

        aql_query, bind_vars = self._build_hybrid_search_query(
            query=query,
            k=k,
            embedding=embedding,
            return_fields=return_fields,
            use_approx=use_approx,
            filter_clause=filter_clause,
            vector_weight=vector_weight,
            keyword_weight=keyword_weight,
            keyword_search_clause=keyword_search_clause,
            metadata_clause=metadata_clause,
        )

        cursor = self.db.aql.execute(aql_query, bind_vars=bind_vars, stream=True)

        results = self._process_search_query(cursor)  # type: ignore

        return results

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> Optional[bool]:
        """Delete by vector ID or other criteria.

        :param ids: List of ids to delete.
        :type ids: Optional[List[str]]
        :param kwargs: Other keyword arguments that can be used to delete vectors.
        :type kwargs: Any
        :return: True if deletion is successful, None if no ids are provided,
            or raises an exception if an error occurs.
        :rtype: Optional[bool]
        """
        if not ids:
            return None

        for result in self.collection.delete_many(ids, **kwargs):  # type: ignore
            if isinstance(result, ArangoServerError):
                raise result

        return True

    def get_by_ids(self, ids: Sequence[str], /) -> list[Document]:
        """Get documents by their IDs.

        :param ids: List of ids to get.
        :type ids: Sequence[str]
        :return: List of Documents with the given ids.
        :rtype: list[Document]
        """
        docs = []
        doc: dict[str, Any]

        for doc in self.collection.get_many(ids):  # type: ignore
            _key = doc.pop("_key")
            page_content = doc.pop(self.text_field)

            docs.append(Document(page_content=page_content, id=_key, metadata=doc))

        return docs

    def max_marginal_relevance_search(
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        return_fields: set[str] = set(),
        use_approx: bool = True,
        embedding: Optional[List[float]] = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Search for documents using Maximal Marginal Relevance (MMR).

        MMR optimizes for both similarity to the query and diversity among the results.
        It helps avoid returning redundant or very similar documents by balancing
        relevance and diversity in the selection process.

        :param query: The text query to search for.
        :type query: str
        :param k: The number of documents to return. Defaults to 4.
        :type k: int
        :param fetch_k: The number of documents to fetch for MMR selection.
            Should be larger than k to allow for diversity selection.
            Defaults to 20.
        :type fetch_k: int
        :param lambda_mult: Controls the diversity vs relevance tradeoff.
            Values between 0 and 1, where 0 = maximum diversity, 1 = maximum relevance.
            Defaults to 0.5.
        :type lambda_mult: float
        :param return_fields: Set of additional document fields to return in results.
            The _key and text fields are always returned.
        :type return_fields: set[str]
        :param use_approx: Whether to use approximate nearest neighbor search.
            Enables faster but potentially less accurate results.
            Defaults to True.
        :type use_approx: bool
        :param embedding: Optional pre-computed embedding for the query.
            If not provided, the query will be embedded using the embedding
            function.
        :type embedding: Optional[List[float]]
        :param kwargs: Additional keyword arguments passed to the search methods.
        :type kwargs: Any
        :return: List of Document objects selected by MMR algorithm.
        :rtype: List[Document]

        .. code-block:: python

            # Search with balanced diversity and relevance
            results = vector_store.max_marginal_relevance_search(
                "machine learning",
                k=3,
                fetch_k=10
            )

            # Emphasize diversity over relevance
            diverse_results = vector_store.max_marginal_relevance_search(
                "neural networks",
                k=5,
                fetch_k=20,
                lambda_mult=0.1  # More diverse results
            )

            # Emphasize relevance over diversity
            relevant_results = vector_store.max_marginal_relevance_search(
                "deep learning",
                k=3,
                fetch_k=15,
                lambda_mult=0.9  # More relevant results
            )
        """
        return_fields.add(self.embedding_field)

        # Embed the query
        query_embedding = embedding or self.embedding.embed_query(query)

        # Fetch the initial documents
        docs = self.similarity_search_by_vector(
            embedding=query_embedding,
            k=fetch_k,
            return_fields=return_fields,
            use_approx=use_approx,
            **kwargs,
        )

        # Get the embeddings for the fetched documents
        embeddings = [doc.metadata[self.embedding_field] for doc in docs]

        # Select documents using maximal marginal relevance
        selected_indices = maximal_marginal_relevance(
            np.array(query_embedding), embeddings, lambda_mult=lambda_mult, k=k
        )

        selected_docs = [docs[i] for i in selected_indices]

        return selected_docs

    def _select_relevance_score_fn(self) -> Callable[[float], float]:
        if self.override_relevance_score_fn is not None:
            return self.override_relevance_score_fn

        # Default strategy is to rely on distance strategy provided
        # in vectorstore constructor
        if self._distance_strategy in [
            DistanceStrategy.COSINE,
            DistanceStrategy.EUCLIDEAN_DISTANCE,
        ]:
            return lambda x: x
        else:
            raise ValueError(
                "No supported normalization function"
                f" for distance_strategy of {self._distance_strategy}."
                "Consider providing relevance_score_fn to ArangoVector constructor."
            )

    @classmethod
    def from_texts(
        cls: Type[ArangoVector],
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[dict]] = None,
        database: Optional[StandardDatabase] = None,
        collection_name: str = "documents",
        search_type: SearchType = DEFAULT_SEARCH_TYPE,
        embedding_field: str = "embedding",
        text_field: str = "text",
        index_name: str = "vector_index",
        distance_strategy: DistanceStrategy = DEFAULT_DISTANCE_STRATEGY,
        num_centroids: int = 1,
        ids: Optional[List[str]] = None,
        overwrite_index: bool = False,
        insert_text: bool = True,
        keyword_index_name: str = "keyword_index",
        keyword_analyzer: str = DEFAULT_ANALYZER,
        rrf_constant: int = DEFAULT_RRF_CONSTANT,
        rrf_search_limit: int = DEFAULT_SEARCH_LIMIT,
        **kwargs: Any,
    ) -> ArangoVector:
        """Create an ArangoVector instance from a list of texts.

        This is a convenience method that creates a new ArangoVector instance,
        embeds the provided texts, and stores them in ArangoDB.

        :param texts: List of text strings to add to the vector store.
        :type texts: List[str]
        :param embedding: The embedding function to use for converting text to vectors.
        :type embedding: langchain.embeddings.base.Embeddings
        :param metadatas: Optional list of metadata dictionaries to associate with each
            text.
        :type metadatas: Optional[List[dict]]
        :param database: The ArangoDB database instance to use.
        :type database: Optional[arango.database.StandardDatabase]
        :param collection_name: The name of the ArangoDB collection to use.
            Defaults to "documents".
        :type collection_name: str
        :param search_type: The type of search to perform.
            Can be either SearchType.VECTOR or SearchType.HYBRID.
            Defaults to SearchType.VECTOR.
        :type search_type: SearchType
        :param embedding_field: The field name to store embeddings. Defaults to
            "embedding".
        :type embedding_field: str
        :param text_field: The field name to store text content. Defaults to "text".
        :type text_field: str
        :param index_name: The name of the vector index. Defaults to "vector_index".
        :type index_name: str
        :param distance_strategy: The distance metric to use. Can be
            DistanceStrategy.COSINE or DistanceStrategy.EUCLIDEAN_DISTANCE.
            Defaults to DistanceStrategy.COSINE.
        :type distance_strategy: DistanceStrategy
        :param num_centroids: Number of centroids for vector index. Defaults to 1.
        :type num_centroids: int
        :param ids: Optional list of unique identifiers for each text.
        :type ids: Optional[List[str]]
        :param overwrite_index: Whether to delete and recreate existing indexes.
            Defaults to False.
        :type overwrite_index: bool
        :param insert_text: Whether to store the text content in the database.
            Required for hybrid search. Defaults to True.
        :type insert_text: bool
        :param keyword_index_name: Name of the keyword search index. Defaults to
            "keyword_index".
        :type keyword_index_name: str
        :param keyword_analyzer: Text analyzer for keyword search.
            Defaults to "text_en".
        :type keyword_analyzer: str
        :param rrf_constant: Constant for RRF scoring in hybrid search. Defaults to 60.
        :type rrf_constant: int
        :param rrf_search_limit: Maximum results for RRF scoring. Defaults to 100.
        :type rrf_search_limit: int
        :param kwargs: Additional keyword arguments passed to the constructor.
        :type kwargs: Any
        :return: A new ArangoVector instance with the texts embedded and stored.
        :rtype: ArangoVector

        .. code-block:: python

            from arango import ArangoClient
            from langchain_arangodb.vectorstores import ArangoVector
            from langchain_community.embeddings import OpenAIEmbeddings

            # Connect to ArangoDB
            client = ArangoClient("http://localhost:8529")
            db = client.db("test", username="root", password="openSesame")

            # Create vector store from texts
            texts = ["hello world", "hello arango", "test document"]
            metadatas = [{"source": "doc1"}, {"source": "doc2"}, {"source": "doc3"}]

            vector_store = ArangoVector.from_texts(
                texts=texts,
                embedding=OpenAIEmbeddings(),
                metadatas=metadatas,
                database=db,
                collection_name="test_collection"
            )

            # Create hybrid search store
            hybrid_store = ArangoVector.from_texts(
                texts=["Machine learning algorithms", "Deep neural networks"],
                embedding=OpenAIEmbeddings(),
                database=db,
                search_type=SearchType.HYBRID,
                collection_name="hybrid_docs",
                overwrite_index=True  # Clean start
            )
        """
        if not database:
            raise ValueError("Database must be provided.")

        if not insert_text and search_type == SearchType.HYBRID:
            raise ValueError("insert_text must be True when search_type is HYBRID")

        embeddings = embedding.embed_documents(list(texts))

        embedding_dimension = len(embeddings[0])

        store = cls(
            embedding,
            embedding_dimension=embedding_dimension,
            database=database,
            collection_name=collection_name,
            search_type=search_type,
            embedding_field=embedding_field,
            text_field=text_field,
            vector_index_name=index_name,
            distance_strategy=distance_strategy,
            num_centroids=num_centroids,
            keyword_index_name=keyword_index_name,
            keyword_analyzer=keyword_analyzer,
            rrf_constant=rrf_constant,
            rrf_search_limit=rrf_search_limit,
            **kwargs,
        )

        if overwrite_index:
            store.delete_vector_index()

            if search_type == SearchType.HYBRID:
                store.delete_keyword_index()

        store.add_embeddings(
            texts, embeddings, metadatas=metadatas, ids=ids, insert_text=insert_text
        )

        return store

    @classmethod
    def from_existing_collection(
        cls: Type[ArangoVector],
        collection_name: str,
        text_properties_to_embed: List[str],
        embedding: Embeddings,
        database: StandardDatabase,
        embedding_field: str = "embedding",
        text_field: str = "text",
        batch_size: int = 1000,
        aql_return_text_query: str = "",
        insert_text: bool = False,
        skip_existing_embeddings: bool = False,
        search_type: SearchType = DEFAULT_SEARCH_TYPE,
        keyword_index_name: str = "keyword_index",
        keyword_analyzer: str = DEFAULT_ANALYZER,
        rrf_constant: int = DEFAULT_RRF_CONSTANT,
        rrf_search_limit: int = DEFAULT_SEARCH_LIMIT,
        **kwargs: Any,
    ) -> ArangoVector:
        """Create an ArangoVector instance from an existing ArangoDB collection.

        This method reads documents from an existing collection, extracts specified
        text properties, embeds them, and creates a new vector store.

        :param collection_name: Name of the existing ArangoDB collection.
        :type collection_name: str
        :param text_properties_to_embed: List of document properties containing text to
            embed. These properties will be concatenated to create
            the text for embedding.
        :type text_properties_to_embed: List[str]
        :param embedding: The embedding function to use for converting text to vectors.
        :type embedding: Embeddings
        :param database: The ArangoDB database instance to use.
        :type database: StandardDatabase
        :param embedding_field: The field name to store embeddings.
            Defaults to "embedding".
        :type embedding_field: str
        :param text_field: The field name to store text content. Defaults to "text".
        :type text_field: str
        :param batch_size: Number of documents to process in each batch.
            Defaults to 1000.
        :type batch_size: int
        :param aql_return_text_query: Custom AQL query to extract text from properties.
            Defaults to "RETURN doc[p]".
        :type aql_return_text_query: str
        :param insert_text: Whether to store the concatenated text in the database.
            Required for hybrid search. Defaults to False.
        :type insert_text: bool
        :param skip_existing_embeddings: Whether to skip documents that already have
            embeddings. Defaults to False.
        :type skip_existing_embeddings: bool
        :param search_type: The type of search to perform.
            Can be either SearchType.VECTOR or SearchType.HYBRID.
            Defaults to SearchType.VECTOR.
        :type search_type: SearchType
        :param keyword_index_name: Name of the keyword search index.
            Defaults to "keyword_index".
        :type keyword_index_name: str
        :param keyword_analyzer: Text analyzer for keyword search.
            Defaults to "text_en".
        :type keyword_analyzer: str
        :param rrf_constant: Constant for RRF scoring in hybrid search. Defaults to 60.
        :type rrf_constant: int
        :param rrf_search_limit: Maximum results for RRF scoring. Defaults to 100.
        :type rrf_search_limit: int
        :param kwargs: Additional keyword arguments passed to the constructor.
        :type kwargs: Any
        :return: A new ArangoVector instance with embeddings created from the
            collection.
        :rtype: ArangoVector
        """
        if not text_properties_to_embed:
            m = "Parameter `text_properties_to_embed` must not be an empty list"
            raise ValueError(m)

        if text_field in text_properties_to_embed:
            m = "Parameter `text_field` must not be in `text_properties_to_embed`"
            raise ValueError(m)

        if not insert_text and search_type == SearchType.HYBRID:
            raise ValueError("insert_text must be True when search_type is HYBRID")

        if not aql_return_text_query:
            aql_return_text_query = "RETURN doc[p]"

        filter_clause = ""
        if skip_existing_embeddings:
            filter_clause = f"FILTER doc.{embedding_field} == null"

        aql_query = f"""
            FOR doc IN @@collection
                {filter_clause}

                LET texts = (
                    FOR p IN @properties
                        FILTER doc[p] != null
                        {aql_return_text_query}
                )

                RETURN {{
                    key: doc._key,
                    text: CONCAT_SEPARATOR(" ", texts),
                }}
        """

        bind_vars = {
            "@collection": collection_name,
            "properties": text_properties_to_embed,
        }

        cursor: Cursor = database.aql.execute(
            aql_query,
            bind_vars=bind_vars,  # type: ignore
            batch_size=batch_size,
            stream=True,
        )

        store: ArangoVector | None = None

        while not cursor.empty():
            batch = cursor.batch()
            batch_list = list(batch)  # type: ignore

            texts = [doc["text"] for doc in batch_list]
            ids = [doc["key"] for doc in batch_list]

            store = cls.from_texts(
                texts=texts,
                embedding=embedding,
                database=database,
                collection_name=collection_name,
                embedding_field=embedding_field,
                text_field=text_field,
                ids=ids,
                insert_text=insert_text,
                search_type=search_type,
                keyword_index_name=keyword_index_name,
                keyword_analyzer=keyword_analyzer,
                rrf_constant=rrf_constant,
                rrf_search_limit=rrf_search_limit,
                **kwargs,
            )

            batch.clear()  # type: ignore

            if cursor.has_more():
                cursor.fetch()

        if store is None:
            raise ValueError(f"No documents found in collection in {collection_name}")

        return store

    def _process_search_query(self, cursor: Cursor) -> List[tuple[Document, float]]:
        data: dict[str, Any]
        score: float
        results = []

        while not cursor.empty():
            for result in cursor:
                data, score, metadata = (
                    result["data"],
                    result["score"],
                    result["metadata"],
                )

                if not data:
                    continue

                _key = data.pop("_key")
                page_content = data.pop(self.text_field)
                doc = Document(
                    page_content=page_content, id=_key, metadata={**data, **metadata}
                )

                results.append((doc, score))

            if cursor.has_more():
                cursor.fetch()

        return results

    def _build_vector_search_query(
        self,
        embedding: List[float],
        k: int,
        return_fields: set[str],
        use_approx: bool,
        filter_clause: str,
        metadata_clause: str,
    ) -> Tuple[str, dict[str, Any]]:
        if self._distance_strategy == DistanceStrategy.COSINE:
            score_func = "APPROX_NEAR_COSINE" if use_approx else "COSINE_SIMILARITY"
            sort_order = "DESC"
        elif self._distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            score_func = "APPROX_NEAR_L2" if use_approx else "L2_DISTANCE"
            sort_order = "ASC"
        else:
            raise ValueError(f"Unsupported metric: {self._distance_strategy}")

        if use_approx:
            if version.parse(self.db.version()) < version.parse("3.12.4"):  # type: ignore
                m = "Approximate Nearest Neighbor search requires ArangoDB >= 3.12.4."
                raise ValueError(m)

            if not self.retrieve_vector_index():
                self.create_vector_index()

        return_fields.update({"_key", self.text_field})
        return_fields_list = list(return_fields)

        aql_query = f"""
            FOR doc IN @@collection
                {filter_clause if not use_approx else ""}
                LET score = {score_func}(doc.{self.embedding_field}, @embedding)
                SORT score {sort_order}
                LIMIT {k}
                {filter_clause if use_approx else ""}
                LET data = KEEP(doc, {return_fields_list})
                LET metadata = {f"({metadata_clause})" if metadata_clause else "{}"}
                RETURN {{data, score, metadata}}
        """

        bind_vars = {
            "@collection": self.collection_name,
            "embedding": embedding,
        }

        return aql_query, bind_vars

    def _build_hybrid_search_query(
        self,
        query: str,
        k: int,
        embedding: List[float],
        return_fields: set[str],
        use_approx: bool,
        filter_clause: str,
        vector_weight: float,
        keyword_weight: float,
        keyword_search_clause: str,
        metadata_clause: str,
    ) -> Tuple[str, dict[str, Any]]:
        """Build the hybrid search query using RRF."""

        if not self.retrieve_keyword_index():
            self.create_keyword_index()

        if self._distance_strategy == DistanceStrategy.COSINE:
            score_func = "APPROX_NEAR_COSINE" if use_approx else "COSINE_SIMILARITY"
            sort_order = "DESC"
        elif self._distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            score_func = "APPROX_NEAR_L2" if use_approx else "L2_DISTANCE"
            sort_order = "ASC"
        else:
            raise ValueError(f"Unsupported metric: {self._distance_strategy}")

        if use_approx:
            if version.parse(self.db.version()) < version.parse("3.12.4"):  # type: ignore
                m = "Approximate Nearest Neighbor search requires ArangoDB >= 3.12.4."
                raise ValueError(m)

            if not self.retrieve_vector_index():
                self.create_vector_index()

        return_fields.update({"_key", self.text_field})
        return_fields_list = list(return_fields)

        if not keyword_search_clause:
            keyword_search_clause = f"""
                SEARCH ANALYZER(
                    doc.{self.text_field} IN TOKENS(@query, @analyzer),
                    @analyzer
                )
            """

        aql_query = f"""
            LET vector_results = (
                FOR doc IN @@collection
                    {filter_clause if not use_approx else ""}
                    LET score = {score_func}(doc.{self.embedding_field}, @embedding)
                    SORT score {sort_order}
                    LIMIT {k}
                    {filter_clause if use_approx else ""}
                    WINDOW {{ preceding: "unbounded", following: 0 }}
                    AGGREGATE rank = COUNT(1)
                    LET rrf_score = {vector_weight} / ({self.rrf_constant} + rank)
                    RETURN {{ key: doc._key, score: rrf_score }}
            )

            LET keyword_results = (
                FOR doc IN @@view
                    {keyword_search_clause}
                    {filter_clause}
                    LET score = BM25(doc)
                    SORT score DESC
                    LIMIT {k}
                    WINDOW {{ preceding: "unbounded", following: 0 }}
                    AGGREGATE rank = COUNT(1)
                    LET rrf_score = {keyword_weight} / ({self.rrf_constant} + rank)
                    RETURN {{ key: doc._key, score: rrf_score }}
            )

            FOR result IN APPEND(vector_results, keyword_results)
                COLLECT key = result.key AGGREGATE score = SUM(result.score)
                SORT score DESC
                LIMIT {self.rrf_search_limit}
                LET data = FIRST(
                    FOR doc IN @@collection
                        FILTER doc._key == key
                        LIMIT 1
                        RETURN KEEP(doc, {return_fields_list})
                )
                LET metadata = {f"({metadata_clause})" if metadata_clause else "{}"}
                RETURN {{ data, score, metadata }}
        """

        bind_vars = {
            "@collection": self.collection_name,
            "@view": self.keyword_index_name,
            "embedding": embedding,
            "query": query,
            "analyzer": self.keyword_analyzer,
        }

        return aql_query, bind_vars



import json
import os
import re
from collections import defaultdict
from math import ceil
from typing import Any, DefaultDict, Dict, List, Optional, Set, Union

import farmhash
import yaml
from arango import ArangoClient
from arango.database import Database, StandardDatabase
from arango.graph import Graph
from langchain_core.embeddings import Embeddings

from langchain_arangodb.graphs.graph_document import (
    Document,
    GraphDocument,
    Node,
    Relationship,
)
from langchain_arangodb.graphs.graph_store import GraphStore

DEFAULT_ANALYZERS = {
    "text_fr",
    "text_pt",
    "identity",
    "text_de",
    "text_zh",
    "text_fi",
    "text_it",
    "text_no",
    "text_nl",
    "text_es",
    "text_ru",
    "text_en",
    "text_sv",
}


def get_arangodb_client(
    url: Optional[str] = None,
    dbname: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> Any:
    """Get the Arango DB client from credentials.

    :param url: Arango DB url. Can be passed in as named arg or set as environment
        var ``ARANGODB_URL``. Defaults to "http://localhost:8529".
    :type url: str
    :param dbname: Arango DB name. Can be passed in as named arg or set as
        environment var ``ARANGODB_DBNAME``. Defaults to "_system".
    :type dbname: str
    :param username: Can be passed in as named arg or set as environment var
        ``ARANGODB_USERNAME``. Defaults to "root".
    :type username: str
    :param password: Can be passed in as named arg or set as environment var
        ``ARANGODB_PASSWORD``. Defaults to "".
    :type password: str

    :return: An arango.database.StandardDatabase.
    :rtype: Any
    :raises ArangoClientError: If the ArangoDB client cannot be created.
    :raises ArangoServerError: If the ArangoDB server cannot be reached.
    :raises ArangoCollectionError: If the collection cannot be created.
    """
    _url: str = url or str(os.environ.get("ARANGODB_URL", "http://localhost:8529"))
    _dbname: str = dbname or str(os.environ.get("ARANGODB_DBNAME", "_system"))
    _username: str = username or str(os.environ.get("ARANGODB_USERNAME", "root"))
    _password: str = password or str(os.environ.get("ARANGODB_PASSWORD", ""))

    return ArangoClient(_url).db(_dbname, _username, _password, verify=True)


class ArangoGraph(GraphStore):
    """ArangoDB wrapper for graph operations.

    :param db: The ArangoDB database instance.
    :type db: StandardDatabase
    :param generate_schema_on_init: Whether to generate the graph schema
        on initialization. Defaults to True.
    :type generate_schema_on_init: bool
    :param schema_sample_ratio: The ratio of documents/edges to sample in relation to
        the Collection size to generate each Collection Schema. If 0, one document/edge
        is used per Collection. Defaults to 0.
    :type schema_sample_ratio: float
    :param schema_graph_name: The name of an existing ArangoDB Graph to specifically
        use to generate the schema. If None, the entire database will be used.
        Defaults to None.
    :type schema_graph_name: Optional[str]
    :param schema_include_examples: Whether to include example values fetched from
        a sample documents as part of the schema. Defaults to True. Lists of size
        higher than **schema_list_limit** will be excluded from the schema, even if
        **schema_include_examples** is set to True. Defaults to True.
    :type schema_include_examples: bool
    :param schema_list_limit: The maximum list size the schema will include as part
        of the example values. If the list is longer than this limit, a string
        describing the list will be used in the schema instead. Default is 32.
    :type schema_list_limit: int
    :param schema_string_limit: The maximum number of characters to include
        in a string. If the string is longer than this limit, a string
        describing the string will be used in the schema instead. Default is 256.
    :type schema_string_limit: int
    :param schema_include_views: Whether to include ArangoDB Views and Analyzers as
        part of the schema passed to the AQL Generation prompt. Default is False.
    :type schema_include_views: bool
        :return: None
        :rtype: None
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.


    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.


    """

    def __init__(
        self,
        db: StandardDatabase,
        generate_schema_on_init: bool = True,
        schema_sample_ratio: float = 0,
        schema_graph_name: Optional[str] = None,
        schema_include_examples: bool = True,
        schema_list_limit: int = 32,
        schema_string_limit: int = 256,
        schema_include_views: bool = False,
    ) -> None:
        """
        Initializes the ArangoGraph instance.

        """
        self.__db: StandardDatabase = db
        self.__async_db = db.begin_async_execution()

        self.__schema = {}
        if generate_schema_on_init:
            self.__schema = self.generate_schema(
                schema_sample_ratio,
                schema_graph_name,
                schema_include_examples,
                schema_list_limit,
                schema_string_limit,
                schema_include_views,
            )

    @property
    def db(self) -> StandardDatabase:
        return self.__db

    @property
    def schema(self) -> Dict[str, Any]:
        """Returns the schema of the Graph Database as a structured object"""
        return self.__schema

    @property
    def get_structured_schema(self) -> Dict[str, Any]:
        """Returns the schema of the Graph Database as a structured object"""
        return self.__schema

    @property
    def schema_json(self) -> str:
        """Returns the schema of the Graph Database as a JSON string

        :return: The schema of the Graph Database as a JSON string
        :rtype: str
        """
        return json.dumps(self.__schema)

    @property
    def schema_yaml(self) -> str:
        """Returns the schema of the Graph Database as a YAML string

        :return: The schema of the Graph Database as a YAML string
        :rtype: str
        """
        return yaml.dump(self.__schema, sort_keys=False)

    def set_schema(self, schema: Dict[str, Any]) -> None:
        """Sets a custom schema for the ArangoDB Database.

        :param schema: The schema to set.
        :type schema: Dict[str, Any]
        :return: None
        :rtype: None
        """
        self.__schema = schema

    def refresh_schema(
        self,
        sample_ratio: float = 0,
        graph_name: Optional[str] = None,
        include_examples: bool = True,
        list_limit: int = 32,
    ) -> None:
        """
        Refresh the graph schema information.

        Parameters:

        :param sample_ratio: A float (0 to 1) to determine the
            ratio of documents/edges sampled in relation to the Collection size
            to generate each Collection Schema. If 0, one document/edge
            is used per Collection. Defaults to 0.
        :type sample_ratio: float
        :param graph_name: The name of an existing ArangoDB Graph to specifically
            use to generate the schema. If None, the entire database will be used.
            Defaults to None.
        :type graph_name: Optional[str]
        :param include_examples: Whether to include example values fetched from
            a sample documents as part of the schema. Defaults to True. Lists of size
            higher than **list_limit** will be excluded from the schema, even if
            **schema_include_examples** is set to True. Defaults to True.
        :type include_examples: bool
        :param list_limit: The maximum list size the schema will include as part
            of the example values. If the list is longer than this limit, a string
            describing the list will be used in the schema instead. Default is 32.
        :type list_limit: int

        :return: None
        :rtype: None
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        self.__schema = self.generate_schema(
            sample_ratio, graph_name, include_examples, list_limit
        )

    def generate_schema(
        self,
        sample_ratio: float = 0,
        graph_name: Optional[str] = None,
        include_examples: bool = True,
        list_limit: int = 32,
        schema_string_limit: int = 256,
        schema_include_views: bool = False,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """
        Generates the schema of the ArangoDB Database and returns it

        :param sample_ratio: A ratio (0 to 1) to determine the
            ratio of documents/edges used (in relation to the Collection size)
            to render each Collection Schema. If 0, one document/edge
            is used per Collection.
        :type sample_ratio: float
        :param graph_name: The name of the graph to use to generate the schema. If
            None, the entire database will be used.
        :type graph_name: Optional[str]
        :param include_examples: A flag whether to scan the database for
            example values and use them in the graph schema. Default is True.
        :type include_examples: bool
        :param list_limit: The maximum number of elements to include in a list.
            If the list is longer than this limit, a string describing the list
            will be used in the schema instead. Default is 32.
        :type list_limit: int
        :param schema_string_limit: The maximum number of characters to include
            in a string. If the string is longer than this limit, a string
            describing the string will be used in the schema instead. Default is 128.
        :type schema_string_limit: int
        :param schema_include_views: Whether to include ArangoDB Views and Analyzers as
            part of the schema passed to the AQL Generation prompt. Default is False.
        :type schema_include_views: bool
        :return: A dictionary containing the graph schema and collection schema.
        :rtype: Dict[str, List[Dict[str, Any]]]
        :raises ValueError: If the sample ratio is not between 0 and 1.
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        if not 0 <= sample_ratio <= 1:
            raise ValueError("**sample_ratio** value must be in between 0 to 1")

        #####
        # Step 1: Generate Graph Schema
        ####

        graph_schema: List[Dict[str, Any]] = []
        if graph_name:
            # Fetch a single graph
            graph: Graph = self.db.graph(graph_name)
            edge_definitions = graph.edge_definitions()

            graph_schema = [{"name": graph_name, "edge_definitions": edge_definitions}]

            # Fetch graph-specific collections
            collection_names = set(graph.vertex_collections())  # type: ignore
            for edge_definition in edge_definitions:  # type: ignore
                collection_names.add(edge_definition["edge_collection"])

        else:
            # Fetch all graphs
            graph_schema = [
                {"graph_name": g["name"], "edge_definitions": g["edge_definitions"]}
                for g in self.db.graphs()  # type: ignore
            ]

            # Fetch all collections
            collection_names = {
                collection["name"]
                for collection in self.db.collections()  # type: ignore
            }

        #####
        # Step 2: Generate Collection Schema
        ####

        # Stores the schema of every ArangoDB Document/Edge collection
        collection_schema: List[Dict[str, Any]] = []
        for collection in self.db.collections():  # type: ignore
            if collection["system"] or collection["name"] not in collection_names:
                continue

            # Extract collection name, type, and size
            col_name: str = collection["name"]
            col_type: str = collection["type"]
            col_size: int = self.db.collection(col_name).count()  # type: ignore

            # Set number of ArangoDB documents/edges to retrieve
            limit_amount = ceil(sample_ratio * col_size) or 1

            aql = f"""
                FOR doc in @@col_name
                    LIMIT {limit_amount}
                    RETURN doc
            """

            doc: Dict[str, Any]
            properties: List[Dict[str, str]] = []
            cursor = self.db.aql.execute(aql, bind_vars={"@col_name": col_name})
            for doc in cursor:  # type: ignore
                for key, value in doc.items():
                    properties.append({key: type(value).__name__})

            collection_schema_entry = {
                "name": col_name,
                "type": col_type,
                "size": col_size,
                "properties": properties,
            }

            if include_examples and col_size > 0:
                collection_schema_entry["example"] = self._sanitize_input(
                    doc, list_limit, schema_string_limit
                )

            collection_schema.append(collection_schema_entry)

        if not schema_include_views:
            return {
                "graph_schema": graph_schema,
                "collection_schema": collection_schema,
            }

        #####
        # Step 3: Generate View Schema
        #####

        view_schema: List[Dict[str, Any]] = []
        for view in self.db.views():  # type: ignore
            view_name = view["name"]
            view_type = view["type"]
            view_info = self.db.view(view_name)
            key = "links" if view_type == "arangosearch" else "indexes"
            view_schema.append(
                {"name": view_name, "type": view_type, key: view_info.get(key, [])}  # type: ignore
            )

        #####
        # Step 4: Generate Analyzer Schema
        #####

        analyzer_schema: List[Dict[str, Any]] = []
        for a in self.db.analyzers():  # type: ignore
            if a["name"] not in DEFAULT_ANALYZERS:
                analyzer_schema.append({a["name"]: a["properties"]})

        return {
            "graph_schema": graph_schema,
            "collection_schema": collection_schema,
            "view_schema": view_schema,
            "analyzer_schema": analyzer_schema,
        }

    def query(
        self,
        query: str,
        params: dict = {},
    ) -> List[Any]:
        """
        Execute an AQL query and return the results.

        :param query: The AQL query to execute.
        :type query: str
        :param params: Additional arguments piped to the function.
            Defaults to None.
        :type params: dict
        :param list_limit: Removes lists above **list_limit** size
            that have been returned from the AQL query.
        :type list_limit: Optional[int]
        :param string_limit: Removes strings above **string_limit** size
            that have been returned from the AQL query.
        :type string_limit: Optional[int]
        :param remaining_params: Remaining params are passed to the AQL query execution.
            Defaults to None.
        :type remaining_params: Optional[dict]

        :return: A list of dictionaries containing the query results.
        :rtype: List[Any]
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        top_k = params.pop("top_k", None)
        list_limit = params.pop("list_limit", 32)
        string_limit = params.pop("string_limit", 256)
        cursor = self.__db.aql.execute(query, **params)

        results = []

        i = 0
        for doc in cursor:  # type: ignore
            results.append(self._sanitize_input(doc, list_limit, string_limit))

            if top_k and i >= top_k:
                break

            i += 1

        return results

    def explain(self, query: str, params: dict = {}) -> List[Dict[str, Any]]:
        """
        Explain an AQL query without executing it.

        :param query: The AQL query to explain.
        :type query: str
        :param params: Additional arguments piped to the function.
            Defaults to None.
        :type params: dict
        :return: A list of dictionaries containing the query explanation.
        :rtype: List[Dict[str, Any]]
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        return self.__db.aql.explain(query)  # type: ignore

    def add_graph_documents(
        self,
        graph_documents: List[GraphDocument],
        include_source: bool = False,
        graph_name: Optional[str] = None,
        update_graph_definition_if_exists: bool = False,
        batch_size: int = 1000,
        use_one_entity_collection: bool = True,
        insert_async: bool = False,
        source_collection_name: Union[str, None] = None,
        source_edge_collection_name: Union[str, None] = None,
        entity_collection_name: Union[str, None] = None,
        entity_edge_collection_name: Union[str, None] = None,
        embeddings: Union[Embeddings, None] = None,
        embedding_field: str = "embedding",
        embed_source: bool = False,
        embed_nodes: bool = False,
        embed_relationships: bool = False,
        capitalization_strategy: str = "none",
    ) -> None:
        """
        Constructs nodes & relationships in the graph based on the
        provided GraphDocument objects.

        :param graph_documents: The GraphDocument objects to add to the graph.
        :type graph_documents: List[GraphDocument]
        :param include_source: Whether to include the source document in the graph.
        :type include_source: bool
        :param graph_name: The name of the graph to add the documents to.
        :type graph_name: Optional[str]
        :param update_graph_definition_if_exists: Whether to update the graph definition
            if it already exists.
        :type update_graph_definition_if_exists: bool
        :param batch_size: The number of documents to process in each batch.
        :type batch_size: int
        :param use_one_entity_collection: Whether to use one entity collection
            for all nodes.
        :type use_one_entity_collection: bool
        :param insert_async: Whether to insert the documents asynchronously.
        :type insert_async: bool
        :param source_collection_name: The name of the source collection.
        :type source_collection_name: Union[str, None]
        :param source_edge_collection_name: The name of the source edge collection.
        :type source_edge_collection_name: Union[str, None]
        :param entity_collection_name: The name of the entity collection.
        :type entity_collection_name: Union[str, None]
        :param entity_edge_collection_name: The name of the entity edge collection.
        :type entity_edge_collection_name: Union[str, None]
        :param embeddings: The embeddings model to use.
        :type embeddings: Union[Embeddings, None]
        :param embedding_field: The field to use for the embedding.
        :type embedding_field: str
        :param embed_source: Whether to embed the source document.
        :type embed_source: bool
        :param embed_nodes: Whether to embed the nodes.
        :type embed_nodes: bool
        :param embed_relationships: Whether to embed the relationships.
        :type embed_relationships: bool
        :param capitalization_strategy: The capitalization strategy to use.
        :type capitalization_strategy: str

        :return: None
        :rtype: None
        :raises ValueError: If the capitalization strategy is not 'lower',
            'upper', or 'none'.
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.

        """
        if not graph_documents:
            return

        if not embeddings and (embed_source or embed_nodes or embed_relationships):
            m = "**embedding** is required to embed source, nodes, or relationships."
            raise ValueError(m)

        def embed_text(text: str) -> list[float]:
            if not embeddings:
                raise ValueError("**embedding** is required to embed text.")

            res: Any = embeddings.embed_documents([text])[0]

            while type(res) is list:
                if type(res[0]) is float:
                    break

                res = res[0]

            return res

        if capitalization_strategy == "none":
            capitalization_fn = lambda x: x  # noqa: E731
        elif capitalization_strategy == "lower":
            capitalization_fn = str.lower
        elif capitalization_strategy == "upper":
            capitalization_fn = str.upper
        else:
            m = "**capitalization_strategy** must be 'lower', 'upper', or 'none'."
            raise ValueError(m)

        #########
        # Setup #
        #########

        suffix = f"{graph_name}_" if graph_name else ""
        if not source_collection_name:
            source_collection_name = f"{suffix}SOURCE"
        if not source_edge_collection_name:
            source_edge_collection_name = f"{suffix}HAS_SOURCE"
        if not entity_collection_name:
            entity_collection_name = f"{suffix}ENTITY"
        if not entity_edge_collection_name:
            entity_edge_collection_name = f"{suffix}LINKS_TO"

        insertion_db = self.__async_db if insert_async else self.__db
        nodes: DefaultDict[str, list[dict[str, Any]]] = defaultdict(list)
        edges: DefaultDict[str, list[dict[str, Any]]] = defaultdict(list)
        edge_definitions_dict: DefaultDict[str, DefaultDict[str, Set[str]]] = (
            defaultdict(lambda: defaultdict(set))
        )

        if include_source:
            self._create_collection(source_collection_name)
            self._create_collection(source_edge_collection_name, is_edge=True)

            from_cols = {entity_collection_name} if use_one_entity_collection else set()

            edge_definitions_dict[source_edge_collection_name][
                "from_vertex_collections"
            ] = from_cols
            edge_definitions_dict[source_edge_collection_name][
                "to_vertex_collections"
            ] = {source_collection_name}

        if use_one_entity_collection:
            self._create_collection(entity_collection_name)
            self._create_collection(entity_edge_collection_name, is_edge=True)

            edge_definitions_dict[entity_edge_collection_name][
                "from_vertex_collections"
            ] = {entity_collection_name}
            edge_definitions_dict[entity_edge_collection_name][
                "to_vertex_collections"
            ] = {entity_collection_name}

        process_node_fn = (
            self._process_node_as_entity
            if use_one_entity_collection
            else self._process_node_as_type
        )

        process_edge_fn = (
            self._process_edge_as_entity
            if use_one_entity_collection
            else self._process_edge_as_type
        )

        #######################
        # Document Processing #
        #######################

        for document in graph_documents:
            source_id_hash = None

            # 1. Process Source Document
            if include_source:
                if not document.source:
                    raise ValueError("Source document is required.")

                source_embedding = (
                    embed_text(document.source.page_content) if embed_source else None
                )

                source_id_hash = self._process_source(
                    document.source,
                    source_collection_name,
                    source_embedding,
                    embedding_field,
                    insertion_db,
                )

            # 2. Process Nodes
            node_key_map = {}
            for i, node in enumerate(document.nodes, 1):
                node.id = str(capitalization_fn(str(node.id)))
                node_key = self._hash(node.id)
                node_key_map[node.id] = node_key

                if embed_nodes:
                    node.properties[embedding_field] = embed_text(node.id)

                node_type = process_node_fn(
                    node_key, node, nodes, entity_collection_name
                )

                # 2.1 Link Source Document to Node
                if include_source:
                    edges[source_edge_collection_name].append(
                        {
                            "_from": f"{node_type}/{node_key}",
                            "_to": f"{source_collection_name}/{source_id_hash}",
                        }
                    )

                    if not use_one_entity_collection:
                        edge_definitions_dict[source_edge_collection_name][
                            "from_vertex_collections"
                        ].add(node_type)

                # 2.2 Batch Insert
                if i % batch_size == 0:
                    self._import_data(insertion_db, nodes, is_edge=False)
                    self._import_data(insertion_db, edges, is_edge=True)

            self._import_data(insertion_db, nodes, is_edge=False)
            self._import_data(insertion_db, edges, is_edge=True)

            # 3. Process Edges
            for i, edge in enumerate(document.relationships, 1):
                source, target = edge.source, edge.target
                source.id = capitalization_fn(str(source.id))
                target.id = capitalization_fn(str(target.id))

                edge_str = f"{source.id} {edge.type} {target.id}"

                if embed_relationships:
                    edge.properties[embedding_field] = embed_text(edge_str)

                if include_source:
                    edge.properties["source_id"] = source_id_hash

                source_key = self._get_node_key(
                    source,
                    nodes,
                    node_key_map,
                    entity_collection_name,
                    process_node_fn,
                )

                target_key = self._get_node_key(
                    target,
                    nodes,
                    node_key_map,
                    entity_collection_name,
                    process_node_fn,
                )

                edge_key = self._hash(edge_str)
                process_edge_fn(
                    edge,
                    edge_str,
                    edge_key,
                    source_key,
                    target_key,
                    edges,
                    entity_collection_name,
                    entity_edge_collection_name,
                    edge_definitions_dict,
                )

                # 3.1 Batch Insert
                if i % batch_size == 0:
                    self._import_data(insertion_db, edges, is_edge=True)
                    self._import_data(insertion_db, nodes, is_edge=False)

            self._import_data(insertion_db, edges, is_edge=True)
            self._import_data(insertion_db, nodes, is_edge=False)

        ##################
        # Graph Creation #
        ##################

        if graph_name:
            edge_definitions: list[dict[str, Union[str, list[str]]]] = [
                {
                    "edge_collection": k,
                    "from_vertex_collections": list(v["from_vertex_collections"]),
                    "to_vertex_collections": list(v["to_vertex_collections"]),
                }
                for k, v in edge_definitions_dict.items()
            ]

            if not self.db.has_graph(graph_name):
                self.db.create_graph(graph_name, edge_definitions)

            elif update_graph_definition_if_exists:
                graph = self.db.graph(graph_name)

                for e_d in edge_definitions:
                    e_c = str(e_d["edge_collection"])
                    if not graph.has_edge_definition(e_c):
                        graph.create_edge_definition(**e_d)  # type: ignore
                    else:
                        graph.replace_edge_definition(**e_d)  # type: ignore

        # Refresh schema after insertions
        self.refresh_schema()

    @classmethod
    def from_db_credentials(
        cls,
        url: Optional[str] = None,
        dbname: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> Any:
        """Convenience constructor that builds Arango DB from credentials.

        :param url: Arango DB url. Can be passed in as named arg or set as environment
                var ``ARANGODB_URL``. Defaults to "http://localhost:8529".
        :type url: str
        :param dbname: Arango DB name. Can be passed in as named arg or set as
                environment var ``ARANGODB_DBNAME``. Defaults to "_system".
        :type dbname: str
        :param username: Can be passed in as named arg or set as environment var
                ``ARANGODB_USERNAME``. Defaults to "root".
        :type username: str
        :param password: Can be passed in as named arg or set as environment var
                ``ARANGODB_USERNAME``. Defaults to "root".
        :type  password: str

        :return: An arango.database.StandardDatabase.
        :rtype: Any
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.

        """
        db = get_arangodb_client(
            url=url, dbname=dbname, username=username, password=password
        )
        return cls(db)

    def _import_data(
        self,
        db: Database,
        data: Dict[str, List[Dict[str, Any]]],
        is_edge: bool,
    ) -> None:
        """Imports data into the ArangoDB database in bulk.

        :param db: The ArangoDB database instance.
        :type db: Database
        :param data: The data to import.
        :type data: Dict[str, List[Dict[str, Any]]]
        :param is_edge: Whether the data is an edge.
        :type is_edge: bool

        :return: None
        :rtype: None
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        for collection, batch in data.items():
            self._create_collection(collection, is_edge)
            db.collection(collection).import_bulk(batch, on_duplicate="update")

        data.clear()

    def _create_collection(
        self, collection_name: str, is_edge: bool = False, **kwargs: Any
    ) -> None:
        """Creates a collection in the ArangoDB database if it does not exist.

        :param collection_name: The name of the collection to create.
        :type collection_name: str
        :param is_edge: Whether the collection is an edge.
        :type is_edge: bool

        :return: None
        :rtype: None
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        if not self.db.has_collection(collection_name):
            self.db.create_collection(collection_name, edge=is_edge, **kwargs)

    def _process_node_as_entity(
        self,
        node_key: str,
        node: Node,
        nodes: DefaultDict[str, list],
        entity_collection_name: str,
    ) -> str:
        """Processes a Graph Document Node into ArangoDB as a unanimous Entity.

        :param node_key: The key of the node.
        :type node_key: str
        :param node: The node to process.
        :type node: Node
        :param nodes: The nodes to process.
        :type nodes: DefaultDict[str, list]
        :param entity_collection_name: The name of the entity collection.
        :type entity_collection_name: str

        :return: The name of the entity collection.
        :rtype: str
        """
        nodes[entity_collection_name].append(
            {
                "_key": node_key,
                "text": node.id,
                "type": node.type,
                **node.properties,
            }
        )
        return entity_collection_name

    def _process_node_as_type(
        self, node_key: str, node: Node, nodes: DefaultDict[str, list], _: str
    ) -> str:
        """Processes a Graph Document Node into ArangoDB based on its Node Type.

        :param node_key: The key of the node.
        :type node_key: str
        :param node: The node to process.
        :type node: Node
        :param nodes: The nodes to process.
        :type nodes: DefaultDict[str, list]
        :param _: The name of the node type.
        :type _: str

        :return: The name of the node type.
        :rtype: str
        """
        node_type = self._sanitize_collection_name(node.type)
        nodes[node_type].append({"_key": node_key, "text": node.id, **node.properties})
        return node_type

    def _process_edge_as_entity(
        self,
        edge: Relationship,
        edge_str: str,
        edge_key: str,
        source_key: str,
        target_key: str,
        edges: DefaultDict[str, list],
        entity_collection_name: str,
        entity_edge_collection_name: str,
        _: DefaultDict[str, DefaultDict[str, set[str]]],
    ) -> None:
        """Processes a Graph Document Edge into ArangoDB as a unanimous Entity.

        :param edge: The edge to process.
        :type edge: Relationship
        :param edge_str: The string representation of the edge.
        :type edge_str: str
        :param edge_key: The key of the edge.
        :type edge_key: str
        :param source_key: The key of the source node.
        :type source_key: str
        :param target_key: The key of the target node.
        :type target_key: str
        :param edges: The edges to process.
        :type edges: DefaultDict[str, list]
        :param entity_collection_name: The name of the entity collection.
        :type entity_collection_name: str
        :param entity_edge_collection_name: The name of the entity edge collection.
        :type entity_edge_collection_name: str
        :param _: The name of the edge type.
        :type _: DefaultDict[str, DefaultDict[str, set[str]]]

        :return: None
        :rtype: None
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.

        """
        edges[entity_edge_collection_name].append(
            {
                "_key": edge_key,
                "_from": f"{entity_collection_name}/{source_key}",
                "_to": f"{entity_collection_name}/{target_key}",
                "type": edge.type,
                "text": edge_str,
                **edge.properties,
            }
        )

    def _process_edge_as_type(
        self,
        edge: Relationship,
        edge_str: str,
        edge_key: str,
        source_key: str,
        target_key: str,
        edges: DefaultDict[str, list],
        _1: str,
        _2: str,
        edge_definitions_dict: DefaultDict[str, DefaultDict[str, set[str]]],
    ) -> None:
        """Processes a Graph Document Edge into ArangoDB based on its Edge Type.

        :param edge: The edge to process.
        :type edge: Relationship
        :param edge_str: The string representation of the edge.
        :type edge_str: str
        :param edge_key: The key of the edge.
        :type edge_key: str
        :param source_key: The key of the source node.
        :type source_key: str
        :param target_key: The key of the target node.
        :type target_key: str
        :param edges: The edges to process.
        :type edges: DefaultDict[str, list]
        :param edge_definitions_dict: The edge definitions dictionary.
        :type edge_definitions_dict: DefaultDict[str, DefaultDict[str, set[str]]]

        :return: None
        :rtype: None
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        source: Node = edge.source
        target: Node = edge.target

        edge_type = self._sanitize_collection_name(edge.type)
        source_type = self._sanitize_collection_name(source.type)
        target_type = self._sanitize_collection_name(target.type)

        edge_definitions_dict[edge_type]["from_vertex_collections"].add(source_type)
        edge_definitions_dict[edge_type]["to_vertex_collections"].add(target_type)

        edges[edge_type].append(
            {
                "_key": edge_key,
                "_from": f"{source_type}/{source_key}",
                "_to": f"{target_type}/{target_key}",
                "text": edge_str,
                **edge.properties,
            }
        )

    def _get_node_key(
        self,
        node: Node,
        nodes: DefaultDict[str, list],
        node_key_map: Dict[str, str],
        entity_collection_name: str,
        process_node_fn: Any,
    ) -> str:
        """Gets the key of a node and processes it if it doesn't exist.

        :param node: The node to process.
        :type node: Node
        :param nodes: The nodes to process.
        :type nodes: DefaultDict[str, list]
        :param node_key_map: The node key map.
        :type node_key_map: Dict[str, str]
        :param entity_collection_name: The name of the entity collection.
        :type entity_collection_name: str
        :param process_node_fn: The function to process the node.
        :type process_node_fn: Any

        :return: The key of the node.
        :rtype: str
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        node.id = str(node.id)
        if node.id in node_key_map:
            return node_key_map[node.id]

        node_key = self._hash(node.id)
        node_key_map[node.id] = node_key
        process_node_fn(node_key, node, nodes, entity_collection_name)

        return node_key

    def _process_source(
        self,
        source: Document,
        source_collection_name: str,
        source_embedding: Union[list[float], None],
        embedding_field: str,
        insertion_db: Database,
    ) -> str:
        """Processes a Graph Document Source into ArangoDB.

        :param source: The source to process.
        :type source: Document
        :param source_collection_name: The name of the source collection.
        :type source_collection_name: str
        :param source_embedding: The embedding of the source.
        :type source_embedding: Union[list[float], None]
        :param embedding_field: The field name to store the embedding.
        :type embedding_field: str
        :param insertion_db: The database to insert the source into.
        :type insertion_db: Database

        :return: The key of the source.
        :rtype: str
        :raises ArangoClientError: If the ArangoDB client cannot be created.
        :raises ArangoServerError: If the ArangoDB server cannot be reached.
        :raises ArangoCollectionError: If the collection cannot be created.
        """
        source_id = self._hash(
            source.id if source.id else source.page_content.encode("utf-8")
        )

        doc = {
            **source.metadata,
            "_key": source_id,
            "text": source.page_content,
            "type": source.type,
        }

        if source_embedding:
            doc[embedding_field] = source_embedding

        insertion_db.collection(source_collection_name).insert(doc, overwrite=True)

        return source_id

    def _hash(self, value: Any) -> str:
        """Applies the Farmhash hash function to a value.

        :param value: The value to hash.
        :type value: Any

        :return: The hashed value.
        :rtype: str
        :raises ValueError: If the value is not a string or has no
            string representation.
        """
        try:
            value_str = str(value)
        except Exception:
            raise ValueError("Value must be a string or have a string representation.")

        return str(farmhash.Fingerprint64(value_str))  # type: ignore

    def _sanitize_collection_name(self, name: str) -> str:
        """
        Modifies a string to adhere to ArangoDB collection name rules.

        - Trims the name to 256 characters if it's too long.
        - Replaces invalid characters with underscores (_).
        - Ensures the name starts with a letter (prepends 'a' if needed).

        :param name: The name to sanitize.
        :type name: str

        :return: The sanitized name.
        :rtype: str
        :raises ValueError: If the collection name is empty.
        """
        if not name:
            raise ValueError("Collection name cannot be empty.")

        name = name[:256]

        # Replace invalid characters with underscores
        name = re.sub(r"[^a-zA-Z0-9_-]", "_", name)

        # Ensure the name starts with a letter; prepend 'a' if not
        if not re.match(r"^[a-zA-Z]", name):
            name = f"Collection_{name}"

        return name

    def _sanitize_input(self, d: Any, list_limit: int, string_limit: int) -> Any:
        """Sanitize the input dictionary or list.

        Sanitizes the input by removing embedding-like values,
        lists with more than **list_limit** elements, that are mostly irrelevant for
        generating answers in a LLM context. These properties, if left in
        results, can occupy significant context space and detract from
        the LLM's performance by introducing unnecessary noise and cost.

        :param d: The input dictionary or list to sanitize.
        :type d: Any
        :param list_limit: The limit for the number of elements in a list.
        :type list_limit: int
        :param string_limit: The limit for the number of characters in a string.
        :type string_limit: int

        :return: The sanitized dictionary or list.
        :rtype: Any
        :raises ValueError: If the input is not a dictionary or list.
        :raises ValueError: If the list limit is less than 0.
        :raises ValueError: If the string limit is less than 0.

        """

        if isinstance(d, dict):
            new_dict = {}
            for key, value in d.items():
                if isinstance(value, dict):
                    sanitized_value = self._sanitize_input(
                        value, list_limit, string_limit
                    )
                    if sanitized_value is not None:
                        new_dict[key] = sanitized_value
                elif isinstance(value, list):
                    if len(value) < list_limit:
                        sanitized_value = self._sanitize_input(
                            value, list_limit, string_limit
                        )
                        if sanitized_value is not None:
                            new_dict[key] = sanitized_value
                    else:
                        new_dict[key] = (
                            f"List of {len(value)} elements of type {type(value[0])}"
                        )
                elif isinstance(value, str):
                    if len(value) > string_limit:
                        new_dict[key] = f"String of {len(value)} characters"
                    else:
                        new_dict[key] = value
                else:
                    new_dict[key] = value
            return new_dict
        elif isinstance(d, list):
            if len(d) == 0:
                return d
            elif len(d) < list_limit:
                arr = []
                for item in d:
                    sanitized_item = self._sanitize_input(
                        item, list_limit, string_limit
                    )
                    if sanitized_item is not None:
                        arr.append(sanitized_item)
                return arr
            else:
                return f"List of {len(d)} elements of type {type(d[0])}"
        else:
            return d



"""Question answering over a graph."""

from __future__ import annotations

import re
from typing import Any, Dict, List, Optional, Tuple

from arango import AQLQueryExecuteError, AQLQueryExplainError
from langchain.chains.base import Chain
from langchain_core.callbacks import CallbackManagerForChainRun
from langchain_core.language_models import BaseLanguageModel
from langchain_core.messages import AIMessage
from langchain_core.prompts import BasePromptTemplate
from langchain_core.runnables import Runnable
from pydantic import Field

from langchain_arangodb.chains.graph_qa.prompts import (
    AQL_FIX_PROMPT,
    AQL_GENERATION_PROMPT,
    AQL_QA_PROMPT,
)
from langchain_arangodb.graphs.graph_store import GraphStore

AQL_WRITE_OPERATIONS: List[str] = [
    "INSERT",
    "UPDATE",
    "REPLACE",
    "REMOVE",
    "UPSERT",
]


class ArangoGraphQAChain(Chain):
    """Chain for question-answering against a graph by generating AQL statements.

    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    graph: GraphStore = Field(exclude=True)
    aql_generation_chain: Runnable[Dict[str, Any], Any]
    aql_fix_chain: Runnable[Dict[str, Any], Any]
    qa_chain: Runnable[Dict[str, Any], Any]
    input_key: str = "query"  #: :meta private:
    output_key: str = "result"  #: :meta private:

    top_k: int = 10
    """Number of results to return from the query"""
    aql_examples: str = ""
    """Specifies the set of AQL Query Examples that promote few-shot-learning"""
    return_aql_query: bool = False
    """ Specify whether to return the AQL Query in the output dictionary"""
    return_aql_result: bool = False
    """Specify whether to return the AQL JSON Result in the output dictionary"""
    max_aql_generation_attempts: int = 3
    """Specify the maximum amount of AQL Generation attempts that should be made"""
    execute_aql_query: bool = True
    """If False, the AQL Query is only explained & returned, not executed"""
    allow_dangerous_requests: bool = False
    """Forced user opt-in to acknowledge that the chain can make dangerous requests."""
    output_list_limit: int = 32
    """Maximum list length to include in the response prompt. Truncated if longer."""
    output_string_limit: int = 256
    """Maximum string length to include in the response prompt. Truncated if longer."""
    force_read_only_query: bool = False
    """If True, the query is checked for write operations and raises an
    error if a write operation is detected."""

    """
    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the chain."""
        super().__init__(**kwargs)
        if self.allow_dangerous_requests is not True:
            raise ValueError(
                "In order to use this chain, you must acknowledge that it can make "
                "dangerous requests by setting `allow_dangerous_requests` to `True`."
                "You must narrowly scope the permissions of the database connection "
                "to only include necessary permissions. Failure to do so may result "
                "in data corruption or loss or reading sensitive data if such data is "
                "present in the database."
                "Only use this chain if you understand the risks and have taken the "
                "necessary precautions. "
                "See https://python.langchain.com/docs/security for more information."
            )

    @property
    def input_keys(self) -> List[str]:
        """Get the input keys for the chain."""
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        """Get the output keys for the chain."""
        return [self.output_key]

    @property
    def _chain_type(self) -> str:
        """Get the chain type."""
        return "graph_aql_chain"

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        *,
        qa_prompt: Optional[BasePromptTemplate] = None,
        aql_generation_prompt: Optional[BasePromptTemplate] = None,
        aql_fix_prompt: Optional[BasePromptTemplate] = None,
        **kwargs: Any,
    ) -> ArangoGraphQAChain:
        """Initialize from LLM.

        :param llm: The language model to use.
        :type llm: BaseLanguageModel
        :param qa_prompt: The prompt to use for the QA chain.
        :type qa_prompt: BasePromptTemplate
        :param aql_generation_prompt: The prompt to use for the AQL generation chain.
        :type aql_generation_prompt: BasePromptTemplate
        :param aql_fix_prompt: The prompt to use for the AQL fix chain.
        :type aql_fix_prompt: BasePromptTemplate
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any
        :return: The initialized ArangoGraphQAChain.
        :rtype: ArangoGraphQAChain
        :raises ValueError: If the LLM is not provided.
        """
        if qa_prompt is None:
            qa_prompt = AQL_QA_PROMPT
        if aql_generation_prompt is None:
            aql_generation_prompt = AQL_GENERATION_PROMPT
        if aql_fix_prompt is None:
            aql_fix_prompt = AQL_FIX_PROMPT

        qa_chain = qa_prompt | llm
        aql_generation_chain = aql_generation_prompt | llm
        aql_fix_chain = aql_fix_prompt | llm

        return cls(
            qa_chain=qa_chain,
            aql_generation_chain=aql_generation_chain,
            aql_fix_chain=aql_fix_chain,
            **kwargs,
        )

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> Dict[str, Any]:
        """
        Generate an AQL statement from user input, use it retrieve a response
        from an ArangoDB Database instance, and respond to the user input
        in natural language.

        Users can modify the following ArangoGraphQAChain Class Variables:

        :param top_k: The maximum number of AQL Query Results to return
        :type top_k: int

        :param aql_examples: A set of AQL Query Examples that are passed to
            the AQL Generation Prompt Template to promote few-shot-learning.
            Defaults to an empty string.
        :type aql_examples: str

        :param return_aql_query: Whether to return the AQL Query in the
            output dictionary. Defaults to False.
        :type return_aql_query: bool

        :param return_aql_result: Whether to return the AQL Query in the
            output dictionary. Defaults to False
        :type return_aql_result: bool

        :param max_aql_generation_attempts: The maximum amount of AQL
            Generation attempts to be made prior to raising the last
            AQL Query Execution Error. Defaults to 3.
        :type max_aql_generation_attempts: int

        :param execute_aql_query: If False, the AQL Query is only
            explained & returned, not executed. Defaults to True.
        :type execute_aql_query: bool

        :param output_list_limit: The maximum list length to display
            in the output. If the list is longer, it will be truncated.
            Defaults to 32.
        :type output_list_limit: int

        :param output_string_limit: The maximum string length to display
            in the output. If the string is longer, it will be truncated.
            Defaults to 256.
        :type output_string_limit: int
        """
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()
        user_input = inputs[self.input_key]

        ######################
        # Generate AQL Query #
        ######################

        aql_generation_output = self.aql_generation_chain.invoke(
            {
                "adb_schema": self.graph.schema_yaml,
                "aql_examples": self.aql_examples,
                "user_input": user_input,
            },
            callbacks=callbacks,
        )

        aql_query = ""
        aql_error = ""
        aql_result = None
        aql_generation_attempt = 1

        aql_execution_func = (
            self.graph.query if self.execute_aql_query else self.graph.explain
        )

        while (
            aql_result is None
            and aql_generation_attempt < self.max_aql_generation_attempts + 1
        ):
            if isinstance(aql_generation_output, str):
                aql_generation_output_content = aql_generation_output
            elif isinstance(aql_generation_output, AIMessage):
                aql_generation_output_content = str(aql_generation_output.content)
            else:
                m = f"Invalid AQL Generation Output: {aql_generation_output} (type: {type(aql_generation_output)})"  # noqa: E501
                raise ValueError(m)

            #####################
            # Extract AQL Query #
            #####################

            pattern = r"```(?i:aql)?(.*?)```"
            matches: List[str] = re.findall(
                pattern, aql_generation_output_content, re.DOTALL
            )

            if not matches:
                _run_manager.on_text(
                    "Invalid Response: ", end="\n", verbose=self.verbose
                )

                _run_manager.on_text(
                    aql_generation_output_content,
                    color="red",
                    end="\n",
                    verbose=self.verbose,
                )

                m = f"Unable to extract AQL Query from response: {aql_generation_output_content}"  # noqa: E501
                raise ValueError(m)

            aql_query = matches[0].strip()

            if self.force_read_only_query:
                is_read_only, write_operation = self._is_read_only_query(aql_query)

                if not is_read_only:
                    error_msg = f"""
                        Security violation: Write operations are not allowed.
                        Detected write operation in query: {write_operation}
                    """
                    raise ValueError(error_msg)

            _run_manager.on_text(
                f"AQL Query ({aql_generation_attempt}):", verbose=self.verbose
            )
            _run_manager.on_text(
                aql_query, color="green", end="\n", verbose=self.verbose
            )

            #############################
            # Execute/Explain AQL Query #
            #############################

            try:
                params = {
                    "top_k": self.top_k,
                    "list_limit": self.output_list_limit,
                    "string_limit": self.output_string_limit,
                }

                aql_result = aql_execution_func(aql_query, params)
            except (AQLQueryExecuteError, AQLQueryExplainError) as e:
                aql_error = str(e.error_message)

                _run_manager.on_text(
                    "AQL Query Execution Error: ", end="\n", verbose=self.verbose
                )
                _run_manager.on_text(
                    aql_error, color="yellow", end="\n\n", verbose=self.verbose
                )

                ########################
                # Retry AQL Generation #
                ########################

                aql_generation_output = self.aql_fix_chain.invoke(
                    {
                        "adb_schema": self.graph.schema_yaml,
                        "aql_query": aql_query,
                        "aql_error": aql_error,
                    },
                    callbacks=callbacks,
                )

            aql_generation_attempt += 1

        if aql_result is None:
            m = f"""
                Maximum amount of AQL Query Generation attempts reached.
                Unable to execute the AQL Query due to the following error:
                {aql_error}
            """
            raise ValueError(m)

        text = "AQL Result:" if self.execute_aql_query else "AQL Explain:"
        _run_manager.on_text(text, end="\n", verbose=self.verbose)
        _run_manager.on_text(
            str(aql_result), color="green", end="\n", verbose=self.verbose
        )

        if not self.execute_aql_query:
            result = {self.output_key: aql_query, "aql_result": aql_result}

            return result

        ########################
        # Interpret AQL Result #
        ########################

        result = self.qa_chain.invoke(  # type: ignore
            {
                "adb_schema": self.graph.schema_yaml,
                "user_input": user_input,
                "aql_query": aql_query,
                "aql_result": aql_result,
            },
            callbacks=callbacks,
        )

        results: Dict[str, Any] = {self.output_key: result}

        if self.return_aql_query:
            results["aql_query"] = aql_generation_output

        if self.return_aql_result:
            results["aql_result"] = aql_result

        return results

    def _is_read_only_query(self, aql_query: str) -> Tuple[bool, Optional[str]]:
        """Check if the AQL query is read-only.

        :param aql_query: The AQL query to check.
        :type aql_query: str

        :return: True if the query is read-only, False otherwise.
        :rtype: Tuple[bool, Optional[str]]
        """
        normalized_query = aql_query.upper()

        for op in AQL_WRITE_OPERATIONS:
            if op in normalized_query:
                return False, op

        return True, None



from typing import Any, Dict, List, Protocol, runtime_checkable

from langchain_arangodb.graphs.graph_document import GraphDocument


@runtime_checkable
class GraphStore(Protocol):
    """Abstract class for graph operations."""

    @property
    def schema_json(self) -> str:
        """Return the schema of the Graph database as JSON."""
        ...

    @property
    def schema_yaml(self) -> str:
        """Return the schema of the Graph database as YAML."""
        ...

    def query(self, query: str, params: dict = {}) -> List[Dict[str, Any]]:
        """Execute a database query."""
        ...

    def explain(self, query: str, params: dict = {}) -> List[Dict[str, Any]]:
        """Explain the query."""
        ...

    def refresh_schema(self) -> None:
        """Refresh the graph schema information."""
        ...

    def add_graph_documents(
        self, graph_documents: List[GraphDocument], include_source: bool = False
    ) -> None:
        """Take GraphDocument as input as uses it to construct a graph."""
        ...



from typing import Any, List, Union

from arango.database import StandardDatabase
from langchain_core.chat_history import BaseChatMessageHistory
from langchain_core.messages import BaseMessage, messages_from_dict


class ArangoChatMessageHistory(BaseChatMessageHistory):
    """Chat message history stored in an ArangoDB database.

    This class provides persistent storage for chat message histories using ArangoDB
    as the backend. It supports session-based message storage with automatic
    collection creation and indexing.

    :param session_id: Unique identifier for the chat session.
    :type session_id: Union[str, int]
    :param db: ArangoDB database instance for storing chat messages.
    :type db: arango.database.StandardDatabase
    :param collection_name: Name of the ArangoDB collection to store messages.
        Defaults to "ChatHistory".
    :type collection_name: str
    :param window: Maximum number of messages to keep in memory (currently unused).
        Defaults to 3.
    :type window: int
    :param args: Additional positional arguments passed to BaseChatMessageHistory.
    :type args: Any
    :param kwargs: Additional keyword arguments passed to BaseChatMessageHistory.
    :type kwargs: Any

    .. code-block:: python

        from arango import ArangoClient
        from langchain_arangodb.chat_message_histories import ArangoChatMessageHistory

        # Connect to ArangoDB
        client = ArangoClient("http://localhost:8529")
        db = client.db("test", username="root", password="openSesame")

        # Create chat message history
        history = ArangoChatMessageHistory(
            session_id="user_123",
            db=db,
            collection_name="chat_sessions"
        )

        # Add messages
        history.add_user_message("Hello! How are you?")
        history.add_ai_message("I'm doing well, thank you!")

        # Retrieve messages
        messages = history.messages
        print(f"Found {len(messages)} messages")

        # Clear session
        history.clear()
    """

    def __init__(
        self,
        session_id: Union[str, int],
        db: StandardDatabase,
        collection_name: str = "ChatHistory",
        window: int = 3,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        # Make sure session id is not null
        if not session_id:
            raise ValueError("Please ensure that the session_id parameter is provided")

        self._session_id = str(session_id)
        self._db = db
        self._collection_name = collection_name
        self._window = window  # TODO: Use this

        if not self._db.has_collection(collection_name):
            self._db.create_collection(collection_name)

        self._collection = self._db.collection(self._collection_name)

        has_index = False
        for index in self._collection.indexes():  # type: ignore
            if "session_id" in index["fields"]:
                has_index = True
                break

        if not has_index:
            self._collection.add_persistent_index(["session_id"], unique=False)

        super().__init__(*args, **kwargs)

    @property
    def messages(self) -> List[BaseMessage]:
        """Retrieve the messages from ArangoDB.

        Retrieves all messages for the current session from the ArangoDB collection,
        sorted by timestamp in descending order (most recent first).

        :return: List of chat messages for the current session,
            sorted in reverse chronological order (most recent first).
        :rtype: List[BaseMessage]

        .. code-block:: python

            # Get all messages for the session
            messages = history.messages
            for msg in messages:
                print(f"{msg.type}: {msg.content}")

            # Check if session has any messages
            if history.messages:
                print(f"Session has {len(history.messages)} messages")
            else:
                print("No messages in this session")
        """
        query = """
            FOR doc IN @@col
                FILTER doc.session_id == @session_id
                SORT doc.time DESC
                RETURN UNSET(doc, ["session_id", "_id", "_rev"])
        """

        bind_vars = {"@col": self._collection_name, "session_id": self._session_id}

        cursor = self._db.aql.execute(query, bind_vars=bind_vars)  # type: ignore

        messages = [
            {"data": {"content": res["content"]}, "type": res["role"]}
            for res in cursor  # type: ignore
        ]

        return messages_from_dict(messages)

    @messages.setter
    def messages(self, messages: List[BaseMessage]) -> None:
        raise NotImplementedError(
            "Direct assignment to 'messages' is not allowed."
            " Use the 'add_messages' instead."
        )

    def add_message(self, message: BaseMessage) -> None:
        """Append the message to the record in ArangoDB.

        Stores a single chat message in the ArangoDB collection associated with
        the current session. The message is stored with its type, content, and
        session identifier.

        :param message: The chat message to add to the history.
        :type message: BaseMessage

        .. code-block:: python

            from langchain_core.messages import HumanMessage, AIMessage

            # Add user message
            user_msg = HumanMessage(content="What is the weather like?")
            history.add_message(user_msg)

            # Add AI response
            ai_msg = AIMessage(content="I don't have access to current weather data.")
            history.add_message(ai_msg)

            # Or use convenience methods
            history.add_user_message("Hello!")
            history.add_ai_message("Hi there!")
        """

        self._db.collection(self._collection_name).insert(
            {
                "role": message.type,
                "content": message.content,
                "session_id": self._session_id,
            },
        )

    def clear(self) -> None:
        """Clear session memory from ArangoDB.

        Removes all messages associated with the current session from the ArangoDB
        collection. The collection itself is preserved for future use.

        .. code-block:: python

            # Add some messages
            history.add_user_message("Hello")
            history.add_ai_message("Hi!")
            print(f"Messages before clear: {len(history.messages)}")  # Output: 2

            # Clear all messages for this session
            history.clear()
            print(f"Messages after clear: {len(history.messages)}")   # Output: 0

            # Collection still exists for future messages
            history.add_user_message("Starting fresh conversation")
            print(f"New message count: {len(history.messages)}")      # Output: 1
        """
        query = """
            FOR doc IN @@col
                FILTER doc.session_id == @session_id
                REMOVE doc IN @@col
        """

        bind_vars = {"@col": self._collection_name, "session_id": self._session_id}

        self._db.aql.execute(query, bind_vars=bind_vars)  # type: ignore



# flake8: noqa
from langchain_core.prompts.prompt import PromptTemplate

AQL_GENERATION_TEMPLATE = """Task: Generate an ArangoDB Query Language (AQL) query from a User Input.

You are an ArangoDB Query Language (AQL) expert responsible for translating a `User Input` into an ArangoDB Query Language (AQL) query.

You are given an `ArangoDB Schema`. It is a YAML Spec containing:
1. `Graph Schema`: Lists all Graphs within the ArangoDB Database Instance, along with their Edge Relationships.
2. `Collection Schema`: Lists all Collections within the ArangoDB Database Instance, along with their document/edge properties and a document/edge example.
3. `View Schema`: Lists all Views within the ArangoDB Database Instance, along with their linked collections and analyzers.
4. `Analyzer Schema`: Lists all custom-built Analyzers within the ArangoDB Database Instance, along with their properties and features. Does not mention the default ArangoDB analyzers (i.e text_en, text_fr, etc.)

You may also be given a set of `AQL Query Examples` to help you create the `AQL Query`. If provided, the `AQL Query Examples` should be used as a reference, similar to how `ArangoDB Schema` should be used.

Things you should do:
- Think step by step.
- Rely on `ArangoDB Schema` and `AQL Query Examples` (if provided) to generate the query.
- Begin the `AQL Query` by the `WITH` AQL keyword to specify all of the ArangoDB Collections required.
- If a `View Schema` is defined and contains analyzers for specific fields, prefer using the View with the `SEARCH` and `ANALYZER` clauses instead of a direct collection scan.
- Use `PHRASE(...)`, `TOKENS(...)`, or `IN TOKENS(...)` as appropriate when analyzers are available on a field.
- Return the `AQL Query` wrapped in 3 backticks (```).
- Use only the provided relationship types and properties in the `ArangoDB Schema` and any `AQL Query Examples` queries.
- Only answer to requests related to generating an AQL Query.
- If a request is unrelated to generating AQL Query, say that you cannot help the user.

Things you should not do:
- Do not use any properties/relationships that can't be inferred from the `ArangoDB Schema` or the `AQL Query Examples`. 
- Do not include any text except the generated AQL Query.
- Do not provide explanations or apologies in your responses.
- Do not generate an AQL Query that removes or deletes any data.

Under no circumstance should you generate an AQL Query that deletes any data whatsoever.

ArangoDB Schema:
{adb_schema}

AQL Query Examples (Optional):
{aql_examples}

User Input:
{user_input}

AQL Query: 
"""

AQL_GENERATION_PROMPT = PromptTemplate(
    input_variables=["adb_schema", "aql_examples", "user_input"],
    template=AQL_GENERATION_TEMPLATE,
)

AQL_FIX_TEMPLATE = """Task: Address the ArangoDB Query Language (AQL) error message of an ArangoDB Query Language query.

You are an ArangoDB Query Language (AQL) expert responsible for correcting the provided `AQL Query` based on the provided `AQL Error`. 

The `AQL Error` explains why the `AQL Query` could not be executed in the database.
The `AQL Error` may also contain the position of the error relative to the total number of lines of the `AQL Query`.
For example, 'error X at position 2:5' denotes that the error X occurs on line 2, column 5 of the `AQL Query`.  

You are also given the `ArangoDB Schema`. It is a YAML Spec containing:
1. `Graph Schema`: Lists all Graphs within the ArangoDB Database Instance, along with their Edge Relationships.
2. `Collection Schema`: Lists all Collections within the ArangoDB Database Instance, along with their document/edge properties and a document/edge example.
3. `View Schema`: Lists all Views within the ArangoDB Database Instance, along with their linked collections and analyzers.
4. `Analyzer Schema`: Lists all custom-built Analyzers within the ArangoDB Database Instance, along with their properties and features. Does not mention the default ArangoDB analyzers (i.e text_en, text_fr, etc.)

You will output the `Corrected AQL Query` wrapped in 3 backticks (```). Do not include any text except the Corrected AQL Query.

Remember to think step by step.

ArangoDB Schema:
{adb_schema}

AQL Query:
{aql_query}

AQL Error:
{aql_error}

Corrected AQL Query:
"""

AQL_FIX_PROMPT = PromptTemplate(
    input_variables=[
        "adb_schema",
        "aql_query",
        "aql_error",
    ],
    template=AQL_FIX_TEMPLATE,
)

AQL_QA_TEMPLATE = """Task: Generate a natural language `Summary` from the results of an ArangoDB Query Language query.

You are an ArangoDB Query Language (AQL) expert responsible for creating a well-written `Summary` from the `User Input` and associated `AQL Result`.

A user has executed an ArangoDB Query Language query, which has returned the AQL Result in JSON format.
You are responsible for creating an `Summary` based on the AQL Result.

You are given the following information:
- `ArangoDB Schema`: contains a schema representation of the user's ArangoDB Database.
- `User Input`: the original question/request of the user, which has been translated into an AQL Query.
- `AQL Query`: the AQL equivalent of the `User Input`, translated by another AI Model. Should you deem it to be incorrect, suggest a different AQL Query.
- `AQL Result`: the JSON output returned by executing the `AQL Query` within the ArangoDB Database.

Remember to think step by step.

Your `Summary` should sound like it is a response to the `User Input`.
Your `Summary` should not include any mention of the `AQL Query` or the `AQL Result`.

ArangoDB Schema:
{adb_schema}

User Input:
{user_input}

AQL Query:
{aql_query}

AQL Result:
{aql_result}

Summary:
"""
AQL_QA_PROMPT = PromptTemplate(
    input_variables=["adb_schema", "user_input", "aql_query", "aql_result"],
    template=AQL_QA_TEMPLATE,
)
